"""
Scorers

Scorers should score a given league setup on a scale from 0 to 1.

The main goal here is that "better league arrangement" gets a higher score.

TODOs:
    * the biggest performance boost we'd get is precomputing team scores and updating that on
    Team.add_player and Team.remove_player. Right now, depth=3 is infeasible past like 100 players.
"""

from typing import TYPE_CHECKING, Dict, Optional

from .utils import MAX_DISTANCE, get_distance


if TYPE_CHECKING:
    from .models import League


# CONVENIENCE SCORERS
def score_practice_day(league: "League") -> float:
    total_matches = 0
    for team in league.teams:
        practice_day = team.practice_day
        matches = sum(
            [1 if practice_day in p.preferred_days else 0 for p in team.players]
        )
        total_matches += matches
    return total_matches / league.size


def score_location(league: "League") -> float:
    score = 1
    league_size = league.size
    for team in league.teams:
        t_lat, t_long = team.latitude, team.longitude
        scaled_distances = [
            min(
                MAX_DISTANCE,
                get_distance(player.latitude, player.longitude, t_lat, t_long),
            )
            / league_size
            for player in team.players
        ]
        score -= sum(scaled_distances)
    return max(0, score)


# PARITY SCORERS
# TODO: make a scorer to ensure even spread of top tier players
def score_size(league: "League") -> float:
    ideal_size = league.ideal_team_size
    squared_errors = [
        (len(team.players) - ideal_size) ** 2 / ideal_size**2 for team in league.teams
    ]
    return max(0, 1 - sum(squared_errors) / len(squared_errors))


def score_grade(league: "League") -> float:
    ideal_grade = league.ideal_team_grade
    squared_errors = [
        (team.get_grade() - ideal_grade) ** 2 / (ideal_grade**2)
        for team in league.teams
    ]
    return max(0, 1 - sum(squared_errors) / len(squared_errors))


def score_skill(league: "League") -> float:
    ideal_skill = league.ideal_team_skill
    squared_errors = [
        (team.get_skill() - ideal_skill) ** 2 / (ideal_skill**2)
        for team in league.teams
    ]
    return max(0, 1 - sum(squared_errors) / len(squared_errors))


def score_elite(league: "League") -> float:
    ideal_elite_players = league.ideal_elite_players
    squared_errors = [
        (team.get_elite_player_count() - ideal_elite_players) ** 2
        / (ideal_elite_players**2)
        for team in league.teams
    ]
    return max(0, 1 - sum(squared_errors) / len(squared_errors))


SCORER_MAP = {
    "skill": score_skill,
    "grade": score_grade,
    "size": score_size,
    "location": score_location,
    "practice_day": score_practice_day,
    "elite": score_elite,
}

# COMPOSITE SCORER

DEFAULT_WEIGHTS = {
    "skill": 0.3,
    "grade": 0.3,
    "size": 0.15,
    "location": 0.05,
    "practice_day": 0.05,
    "elite": 0.15,
}


def score_league(
    league: "League", weights: Optional[Dict[str, float]] = None, verbose=False
) -> float:
    if weights is None:
        weights = DEFAULT_WEIGHTS
    score: float = 0
    total_weight: float = 0
    s = ""
    for scorer_key, weight in weights.items():
        scorer = SCORER_MAP[scorer_key]
        score += weight * scorer(league)
        total_weight += weight
        if verbose:
            s += f"{scorer(league):.3f} "
    if verbose:
        print(s, list(weights.keys()))
    return score / total_weight



def get_distance(lat1, long1, lat2, long2):
    return ((lat1 - lat2) ** 2 + (long1 - long2) ** 2) ** 0.5


practice_fields = [  # latitude longitude
    (42.38980588369338, -71.13301159861518),  # danehy
    (42.37292538868882, -71.08361099742957),  # gsm
    (42.376852476244125, -71.12084010564928),  # common
]

# Maximum distance between fields. Natural length scale for scoring
MAX_DISTANCE = 0.052205081257465445
CENTROID_LAT = sum([field[0] for field in practice_fields]) / len(practice_fields)
CENTROID_LONG = sum([field[1] for field in practice_fields]) / len(practice_fields)



from dataclasses import asdict, dataclass, field
from typing import List, Optional, Set

import pandas as pd

from .constraints import breaks_practice_constraint
from .scorers import SCORER_MAP
from .utils import CENTROID_LAT, CENTROID_LONG


ELITE_PLAYER_SKILL_LEVEL = 1


@dataclass(frozen=True)
class Player:
    id: int
    first_name: str
    last_name: str
    grade: int
    skill: int
    coach_skill: int
    parent_skill: int
    unavailable_days: str
    preferred_days: str
    latitude: float
    longitude: float
    frozen: bool
    school: str
    comment: str

    @classmethod
    def from_raw_dict(cls, raw_dict):
        """Preprocess a raw csv row in dictionary form to create a Player."""
        if pd.isnull(raw_dict["unavailable_days"]):
            raw_dict["unavailable_days"] = ""
        if pd.isnull(raw_dict["preferred_days"]):
            raw_dict["preferred_days"] = ""
        if pd.isnull(raw_dict["latitude"]):
            raw_dict["latitude"] = CENTROID_LAT
        if pd.isnull(raw_dict["longitude"]):
            raw_dict["longitude"] = CENTROID_LONG

        raw_dict["skill"] = raw_dict["coach_skill"]
        if pd.isnull(raw_dict["skill"]):
            raw_dict["skill"] = raw_dict["parent_skill"]

        return cls(**raw_dict)

    def to_raw_dict(self) -> dict:
        """Create a raw dict that can be saved to a csv."""
        raw_dict = asdict(self)
        del raw_dict["skill"]
        if raw_dict["latitude"] == CENTROID_LAT:
            raw_dict["latitude"] = pd.NA
        if raw_dict["longitude"] == CENTROID_LONG:
            raw_dict["longitude"] = pd.NA
        return raw_dict

    def __post_init__(self):
        """Validate player data."""
        for day in self.unavailable_days:
            if day in self.preferred_days:
                raise ValueError(
                    f"Failed to create player {self.first_name} {self.last_name} ({self.id}) "
                    f"due to invalid day preferences: {day} was marked as both unavailable "
                    f"({self.unavailable_days}) and preferred ({self.preferred_days}). "
                    "Please correct this in the csv and retry."
                )


@dataclass
class Team:
    name: str
    practice_day: str
    latitude: float
    longitude: float
    players: Set[Player] = field(default_factory=set)
    # TODO: practice time

    def add_player(self, player: Player) -> None:
        self.players.add(player)

    def remove_player(self, player: Player) -> None:
        self.players.remove(player)

    def get_skill(self) -> float:
        if len(self.players) == 0:
            return 0
        return sum([player.skill for player in self.players]) / len(self.players)

    def get_grade(self) -> float:
        if len(self.players) == 0:
            return 0
        return sum([player.grade for player in self.players]) / len(self.players)

    def get_elite_player_count(self) -> int:
        return [p.skill for p in self.players].count(ELITE_PLAYER_SKILL_LEVEL)

    def __repr__(self):
        return (
            f"Team {self.name:11} "
            f"({self.practice_day} "
            f"size={len(self.players)}, "
            f"skill={self.get_skill():.2f}, "
            f"grade={self.get_grade():.2f})"
        )


@dataclass(frozen=True)
class Move:
    """A class representing moving a player from one team to another."""

    player: Player
    team_from: Optional[Team] = None
    team_to: Optional[Team] = None

    def __repr__(self):
        return (
            f"Move {self.player}\n"
            f"\t from: {self.team_from}\n"
            f"\t   to: {self.team_to}\n"
        )


@dataclass
class League:
    teams: List[Team]
    available_players: Set[Player]  # TODO: consider a PQ

    def __post_init__(self):
        self.ideal_team_grade = sum([p.grade for p in self.players]) / len(self.players)
        self.ideal_team_skill = sum([p.skill for p in self.players]) / len(self.players)
        self.ideal_team_size = len(self.players) / len(self.teams)
        self.ideal_elite_players = [p.skill for p in self.players].count(
            ELITE_PLAYER_SKILL_LEVEL
        ) / len(self.teams)
        self.size = len(self.players)

    @property
    def players(self) -> List[Player]:
        """Return all players in the league."""
        return [p for team in self.teams for p in team.players] + list(
            self.available_players
        )

    def get_next_player(self) -> Player:
        return max(self.available_players, key=lambda p: p.skill)

    def apply_moves(self, moves: List[Move]) -> None:
        for move in moves:
            if move.team_from is not None:
                move.team_from.remove_player(move.player)
            else:
                self.available_players.remove(move.player)
            if move.team_to is not None:
                move.team_to.add_player(move.player)
            else:
                self.available_players.add(move.player)

    def undo_moves(self, moves: List[Move]) -> None:
        for move in moves:
            if move.team_to is not None:
                move.team_to.remove_player(move.player)
            else:
                self.available_players.remove(move.player)
            if move.team_from is not None:
                move.team_from.add_player(move.player)
            else:
                self.available_players.add(move.player)

    def reset_league(self) -> None:
        for team in self.teams:
            for player in list(team.players):
                self.apply_moves([Move(player=player, team_from=team, team_to=None)])

    def details(self) -> None:
        score_info = {}
        for key, scorer in SCORER_MAP.items():
            score_info[f"{key}_score"] = scorer(self)
        with pd.option_context("display.precision", 3):
            print(pd.DataFrame([score_info]))

        team_info_dicts = []
        for team in self.teams:
            team_info_dict = {
                "name": team.name,
                "practice_day": team.practice_day,
                "size": len(team.players),
                "first_round_picks": team.get_elite_player_count(),
                "mean_skill": team.get_skill(),
                "mean_grade": team.get_grade(),
            }

            team_info_dicts.append(team_info_dict)
        team_info_df = pd.DataFrame.from_records(team_info_dicts)
        with pd.option_context("display.precision", 3):
            print(team_info_df)

    @classmethod
    def from_csvs(cls, player_csv: str, team_csv: str) -> "League":
        # TODO: check constraints here
        player_info = pd.read_csv(player_csv)
        team_info = pd.read_csv(team_csv)
        teams = {}
        available_players = set()
        for i, row in team_info.iterrows():
            teams[row["name"]] = Team(**row.to_dict())

        for i, row in player_info.iterrows():
            player_dict = row.to_dict()
            team = player_dict.pop("team", None)
            player = Player.from_raw_dict(player_dict)
            if not pd.isnull(team):
                move = Move(player=player, team_from=None, team_to=teams[team])
                if breaks_practice_constraint(move):
                    raise ValueError(
                        f"Failed to add {player} to Team {team}. The team's practice day "
                        f"({teams[team].practice_day}) is in the players unavailable days "
                        f"({player.unavailable_days}). Please correct this in the csv and retry."
                    )
                teams[team].add_player(player)
            else:
                available_players.add(player)

        return cls(teams=list(teams.values()), available_players=available_players)

    def to_csvs(self, player_csv: str, team_csv: str) -> None:
        teams = []
        players = []
        for team in self.teams:
            team_dict = asdict(team)
            del team_dict["players"]
            teams.append(team_dict)
            for player in team.players:
                players.append(asdict(player) | {"team": team.name})

        for player in self.available_players:
            players.append(asdict(player))

        print(f"Saving player information to {player_csv}")
        pd.DataFrame.from_records(players).to_csv(player_csv, index=False)

        print(f"Saving team information to {team_csv}")
        pd.DataFrame.from_records(teams).to_csv(team_csv, index=False)

    def __repr__(self):
        s = f"{len(self.teams)} Teams:\n"
        for team in self.teams:
            s += f"{team}\n"
        s += f"{len(self.available_players)} players available for assignment"
        return s



import argparse
import logging

from geopy.extra.rate_limiter import RateLimiter
from geopy.geocoders import Nominatim
import numpy as np
import pandas as pd
from tqdm import tqdm


# Prepare location finding
geolocator = Nominatim(user_agent="cyslf")
geocode = RateLimiter(geolocator.geocode, min_delay_seconds=1 / 2)

tqdm.pandas()

pd.set_option("display.max_rows", None)

parser = argparse.ArgumentParser(
    description="Process raw registration forms into a standardized csv."
)
parser.add_argument(
    "--division",
    "--div",
    "-d",
    type=str,
    help="division without season (e.g. 'Boys Grades 3-4'). Spelling needs to match form data.",
)
parser.add_argument(
    "--old_registration",
    "--old_reg",
    type=str,
    help="Old registration csv. Where to pull coach evals + past teams from.",
)
parser.add_argument(
    "--registration", "--reg", type=str, help="Current registration csv."
)
parser.add_argument(
    "--output_stem",
    "-o",
    type=str,
    help="output csv stem. write outputs to '{stem}-players.csv'",
)


def _load_existing_player_data(filename: str, division: str):
    print("=====")
    print(f"Reading {filename}.")
    existing_players_raw = pd.read_csv(filename)
    # Only keep teams if the players were in the relevant division.
    # This is because a Spring 2nd-grader on Red shouldn't stay on Red, but
    # we do want to know their skill score.
    in_division = existing_players_raw["Division"].str.contains(division)
    existing_players_raw.loc[~in_division, "Assigned Team"] = np.nan

    column_map = {
        "Player rating - Effectiveness": "coach_skill",
        "Lastname": "last_name",
        "Firstname": "first_name",
        "Assigned Team": "team",
    }
    existing_players = existing_players_raw.rename(columns=column_map)[
        column_map.values()
    ]

    # Extract skill
    existing_players["coach_skill"] = (
        existing_players["coach_skill"]
        .str.extract("(\d+)", expand=False)  # noqa: W605
        .astype(float)
    )

    # Construct name for matching to current registration.
    full_name = existing_players["first_name"] + existing_players["last_name"]
    normalized_name = full_name.str.lower().str.replace("[^a-zA-Z]", "", regex=True)
    existing_players["name_key"] = normalized_name
    existing_players = (
        existing_players.drop(columns=["first_name", "last_name"])
        # there seemed to be duplicate rows (with missing skill scores) so
        # let's take the highest score for each name. Note this assumes
        # that kids have unique name keys
        .sort_values(by="coach_skill").drop_duplicates("name_key")
    )

    print(f"Loaded {len(existing_players)} existing players for skill/team lookup.")
    print(f"{in_division.sum()} were found in division: {division}")
    print(existing_players.head())
    print("=====")
    return existing_players


def _lookup_location(address):
    # If an address is poorly formed, geopy gives sad-looking warnings,
    # so let's temporarily disable this.
    logging.getLogger("geopy").setLevel(logging.ERROR)
    location = geocode(address)
    logging.getLogger("geopy").setLevel(logging.WARNING)
    if location:
        return location.latitude, location.longitude
    else:
        print(f"Failed to find address: {address}")
        return np.nan, np.nan


def _load_registration_data(filename):
    print("=====")
    print(f"Reading {filename}")
    registrations_raw = pd.read_csv(filename)

    # Join the two school columns
    has_other_school = ~pd.isnull(registrations_raw["School Name other:"])
    registrations_raw.loc[has_other_school, "School Name"] = registrations_raw[
        has_other_school
    ]["School Name other:"]

    # Look up player latitude / longitude
    print("Looking up latitude/longitude from addresses.")
    registrations_raw["Postal Code"] = registrations_raw["Postal Code"].astype(str)
    registrations_raw["Address"] = registrations_raw[
        ["Street", "City", "Region", "Postal Code"]
    ].agg(", ".join, axis=1)
    registrations_raw["Location"] = registrations_raw["Address"].progress_apply(
        _lookup_location
    )
    registrations_raw["latitude"] = registrations_raw["Location"].apply(lambda x: x[0])
    registrations_raw["longitude"] = registrations_raw["Location"].apply(lambda x: x[1])

    column_map = {
        "Player Last Name": "last_name",
        "Player First Name": "first_name",
        "Current Grade": "grade",
        "Parental assessment of player ability/athleticism:": "parent_skill",
        "School Name": "school",
        "Special Requests": "comment",
        "longitude": "longitude",
        "latitude": "latitude",
        "Teammate Request": "requested_teammate",
    }
    registrations = registrations_raw.rename(columns=column_map)[column_map.values()]

    # Extract grade
    registrations["grade"] = (
        registrations["grade"]
        .str.extract("(\d+)", expand=False)  # noqa: W605
        .astype(int)
    )

    # Extract skill
    registrations["parent_skill"] = (
        registrations["parent_skill"]
        .str.extract("(\d+)", expand=False)  # noqa: W605
        .astype(float)
    )
    registrations["parent_skill"] = (
        11 - registrations["parent_skill"]
    )  # These have the opposite order of coach skill, so invert.

    # Construct name for matching to current registration.
    full_name = registrations["first_name"] + registrations["last_name"]
    normalized_name = full_name.str.lower().str.replace("[^a-zA-Z]", "", regex=True)
    registrations["name_key"] = normalized_name

    print(f"Found {len(registrations)} registrations")
    print(registrations.head())
    print("=====")
    return registrations


def _merge_existing_data_with_registrations(existing_players, registrations):
    print("=====")
    print("Integrating old data with registrations.")
    existing_player_match = registrations["name_key"].isin(existing_players["name_key"])
    print(
        f"{existing_player_match.sum()} out of {len(registrations)} players were matched to "
        "existing player data."
    )
    print("The following players were NOT matched:")
    print(registrations[~existing_player_match])

    ordered_columns = [
        "id",
        "last_name",
        "first_name",
        "grade",
        "team",
        "coach_skill",
        "parent_skill",
        "longitude",
        "latitude",
        "preferred_days",
        "unavailable_days",
        "frozen",
        "school",
        "comment",
    ]
    players = registrations.merge(existing_players, how="left", on="name_key")
    players["id"] = players.index.values

    # Freeze players to a team if they already have one
    players["frozen"] = True
    missing_team = pd.isnull(players.team)
    players.loc[missing_team, "frozen"] = False

    # Fill missing columns with nans
    for col in ordered_columns:
        if col not in players.columns:
            players[col] = np.nan

    players = players[ordered_columns]

    print(f"Finished merging old data into registrations. Total # of players: {len(players)}")
    print(players.head())
    print("=====")
    return players


def _validate_players(players):
    print("=====")
    print("Running verification checks:")

    full_name = players["first_name"] + players["last_name"]
    normalized_name = full_name.str.lower().str.replace("[^a-zA-Z]", "", regex=True)
    name_counts = normalized_name.value_counts()
    duplicate_names = name_counts[name_counts > 1].index.values
    has_duplicate = normalized_name.isin(duplicate_names)
    if has_duplicate.sum() > 0:
        print(
            f"{has_duplicate.sum()} rows have names that match other rows, so please check if "
            "they're duplicates:"
        )
        print(
            players[has_duplicate].sort_values(by="last_name")[
                ["id", "first_name", "last_name"]
            ]
        )

    missing_last_name = pd.isnull(players.last_name)
    if missing_last_name.sum() > 0:
        print(f"{missing_last_name.sum()} players are missing a last name:")
        print(players[missing_last_name][["id", "first_name", "last_name"]])
        print()

    missing_first_name = pd.isnull(players.first_name)
    if missing_first_name.sum() > 0:
        print(f"{missing_first_name.sum()} players are missing a first name:")
        print(players[missing_first_name][["id", "first_name", "last_name"]])
        print()

    missing_grade = pd.isnull(players.grade)
    if missing_grade.sum() > 0:
        print(f"{missing_grade.sum()} players are missing a grade:")
        print(players[missing_grade][["id", "first_name", "last_name", "grade"]])
        print()

    missing_skill = pd.isnull(players.coach_skill) & pd.isnull(players.parent_skill)
    if missing_skill.sum() > 0:
        print(f"{missing_skill.sum()} players don't have a coach or parent skill:")
        print(
            players[missing_skill][
                ["id", "first_name", "last_name", "coach_skill", "parent_skill"]
            ]
        )
        print()

        # For testing let's give them the worst score
        # I should ask Jason about this though
        players.loc[missing_skill, "parent_skill"] = 10

    print("Please manually fix any listed problems before making teams.^")


def main():
    args = parser.parse_args()

    # Pull team and coach score from spring 2022
    existing_players = _load_existing_player_data(args.old_registration, args.division)

    # Load current registrations
    registrations = _load_registration_data(args.registration)

    # Merge the old data into the current registration data.
    players = _merge_existing_data_with_registrations(existing_players, registrations)

    # Validate the player list
    _validate_players(players)

    print("=====")
    player_outfile = f"{args.output_stem}-players.csv"
    print(f"Saving to {player_outfile}")
    players.to_csv(player_outfile, index=False)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cyslf']

package_data = \
{'': ['*']}

install_requires = \
['geopy>=2.2.0,<3.0.0',
 'numpy>=1.23.1,<2.0.0',
 'pandas>=1.4.3,<2.0.0',
 'tqdm>=4.64.0,<5.0.0']

entry_points = \
{'console_scripts': ['make-teams = cyslf.make_teams:main',
                     'prepare-player-data = cyslf.prepare_player_data:main']}

setup_kwargs = {
    'name': 'cyslf',
    'version': '0.1.0',
    'description': 'CYS League Formation',
    'long_description': None,
    'author': 'Evan Tey',
    'author_email': 'evantey14@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import itertools as it
from typing import Dict, List, Optional

from .constraints import breaks_constraints
from .models import League, Move, Player
from .scorers import score_league


def find_best_moves(
    player: Player,
    league: League,
    depth: int = 1,
    weights: Optional[Dict[str, float]] = None,
) -> List[Move]:
    queue = []
    best_moves = []
    best_score: float = -1
    old_team = None
    for team in league.teams:
        if team == old_team:
            continue
    for team in league.teams:
        queue.append([Move(player=player, team_from=old_team, team_to=team)])

    while len(queue) > 0:
        proposed_moves = queue.pop()
        if breaks_constraints(proposed_moves):
            continue
        league.apply_moves(proposed_moves)
        score = score_league(league, weights=weights)
        league.undo_moves(proposed_moves)
        if score > best_score:
            best_score = score
            best_moves = proposed_moves

        if len(proposed_moves) < depth:
            last_team = proposed_moves[-1].team_to
            if last_team is None:
                raise ValueError(
                    f"Proposed move tried to unassign a player, which is not allowed. Please check "
                    f"move proposal creation. Move: {proposed_moves[-1]}"
                )

            for p, t in it.product(last_team.players, league.teams):
                moved_players = [move.player for move in proposed_moves]
                if p in moved_players or t.name == last_team.name:
                    continue
                queue.append(
                    [*proposed_moves, Move(player=p, team_from=last_team, team_to=t)]
                )
    return best_moves



import argparse
import configparser

from tqdm import tqdm

from cyslf.models import League
from cyslf.optimize import find_best_moves
from cyslf.scorers import DEFAULT_WEIGHTS


parser = argparse.ArgumentParser(description="Assign players to teams")
parser.add_argument(
    "--input_stem",
    "-i",
    type=str,
    help="input csv stem. read '{stem}-players.csv' and '{stem}-teams.csv'",
)
parser.add_argument(
    "--output_stem",
    "-o",
    type=str,
    help="output csv stem. write outputs to '{stem}-players.csv' and '{stem}-teams.csv'",
)
parser.add_argument(
    "--config",
    "-c",
    type=str,
    help="config file with scorer weights.",
)


def main():
    args = parser.parse_args()

    player_input_file = f"{args.input_stem}-players.csv"
    team_input_file = f"{args.input_stem}-teams.csv"
    print(
        f"Loading league from {player_input_file} (players) and {team_input_file} (teams)"
    )
    league = League.from_csvs(player_input_file, team_input_file)

    weights = None
    if args.config is not None:
        config = configparser.ConfigParser()
        config.read(args.config)
        weights = {}
        for key in config["weights"].keys():
            weights[key] = float(config["weights"][key])
    else:
        weights = DEFAULT_WEIGHTS
    print(f"Using weights {weights}.")

    print(f"Starting to assign {len(league.available_players)} available players.")
    for i in tqdm(range(len(league.available_players))):
        player = league.get_next_player()
        best_moves = find_best_moves(player, league, depth=2, weights=weights)
        league.apply_moves(best_moves)
    league.details()

    # TODO: add override check
    player_output_file = f"{args.output_stem}-players.csv"
    team_output_file = f"{args.output_stem}-teams.csv"

    league.to_csvs(player_output_file, team_output_file)


if __name__ == "__main__":
    main()
